/*
 * Copyright 2016 - 2025 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#include <imgui.h>

#include <studio/imguiutil.hpp>

#include "projectexplorer.hpp"
#include "projecttreemodel.hpp"

namespace studio {

ProjectTreeModel::ProjectTreeModel(
	ProjectExplorer &explorer,
	ox::StringParam name,
	ProjectTreeModel *parent) noexcept:
	m_explorer(explorer),
	m_parent(parent),
	m_name(std::move(name)) {
}

ProjectTreeModel::ProjectTreeModel(ProjectTreeModel &&other) noexcept:
	m_explorer(other.m_explorer),
	m_parent(other.m_parent),
	m_name(std::move(other.m_name)),
	m_children(std::move(other.m_children)) {
}

void ProjectTreeModel::draw(turbine::Context &tctx) const noexcept {
	constexpr auto dirFlags = ImGuiTreeNodeFlags_OpenOnArrow | ImGuiTreeNodeFlags_OpenOnDoubleClick;
	if (!m_children.empty()) {
		if (ImGui::TreeNodeEx(m_name.c_str(), dirFlags)) {
            drawDirContextMenu();
			for (auto const&child : m_children) {
				child->draw(tctx);
			}
			ImGui::TreePop();
		} else {
			ig::IDStackItem const idStackItem{m_name};
            drawDirContextMenu();
		}
	} else {
		auto const path = fullPath();
		auto const name = ox::sfmt<ox::BasicString<255>>("{}##{}", m_name, path);
		if (ImGui::TreeNodeEx(name.c_str(), ImGuiTreeNodeFlags_Leaf)) {
			if (ImGui::IsItemHovered() && ImGui::IsMouseDoubleClicked(0)) {
				m_explorer.fileChosen.emit(path);
			}
			if (ImGui::BeginPopupContextItem("FileMenu", ImGuiPopupFlags_MouseButtonRight)) {
				if (ImGui::MenuItem("Delete")) {
					m_explorer.deleteItem.emit(path);
				}
				ImGui::EndPopup();
			}
			ImGui::TreePop();
			ig::dragDropSource([this] {
				ImGui::Text("%s", m_name.c_str());
			});
		}
	}
}

void ProjectTreeModel::setChildren(ox::Vector<ox::UPtr<ProjectTreeModel>> children) noexcept {
	m_children = std::move(children);
}

void ProjectTreeModel::drawDirContextMenu() const noexcept {
    if (ImGui::BeginPopupContextItem("DirMenu", ImGuiPopupFlags_MouseButtonRight)) {
        if (ImGui::MenuItem("Add Item")) {
            m_explorer.addItem.emit(fullPath());
        }
        if (ImGui::MenuItem("Add Directory")) {
            m_explorer.addDir.emit(fullPath());
        }
		if (ImGui::MenuItem("Delete")) {
			m_explorer.deleteItem.emit(fullPath());
		}
        ImGui::EndPopup();
    }
}

ox::BasicString<255> ProjectTreeModel::fullPath() const noexcept {
	if (m_parent) {
		return m_parent->fullPath() + "/" + m_name;
	}
	return {};
}

}
